"""
Heckweasel: Rename GnuImp

"""

import argparse
import datetime
import logging
import os
import pprint
import re
import shutil
import sys
import traceback
import yaml

import tqdm # type: ignore

from pathlib import Path
from typing import Sequence, Union

from PIL import Image

from . import (ANNULUS,
               LOGO,
               __version__)
from .gnuimpregexp import (CONTENTS_EXCEPTIONS,
                           CONTENTS_REGEXP,
                           DEBUG_REGEXP,
                           FILENAME_EXCEPTIONS,
                           FILENAME_REGEXP,
                           ICON_REGEXP,
                           LOGO_REGEXP,
                           MASCOT_REGEXP,
                           MASCOT_SCALABLE_REGEXP,
                           SCALABLE_REGEXP,
                           SKIP_PROCESSING,
                           SPLASH_REGEXP,
                           MAKE_DIRECTORIES)
from .misc import (TqdmLoggingHandler,
                   is_binary,
                   find_file,
                   reverse_domain,
                   PROBABLY_BINARY_EXTENSIONS
                   )
from .regexptools import (Replacement,
                          apply_regexps,
                          one_matches,
                          process_text_file)


logger = logging.getLogger(__name__)

####
# TODO
#
# ** wilber.ico and wilber.bmp in windows [we can make these from the input pngs and output them as mascot.bmp/ico]
# * README - need to ttalk about canvas shadow, formats and styles for icon.s etc.

####
# Later TODO
#
# * Add a 'additional files' config, like for installing brushes and such as part of the package. We'd have to patch
#   the appropirate meson files to do it.
# ** we might want to allow color vs. symbolic icons
# * impliment versioning somehow ? fill in appropirate version variables as needed.
# * Fixing various urls, dialogs, etc. (also the text wiggle in about, basically sort of bury the original stuff)
# * We should fix the README, like make a README generated
# * We should fix help menu
# * We should fix references to URLs in the software itself in various places
# * It'd be cool to have a way to apply certain regex by filename instead of globally
# * probably disable check for updates too
# * fix script-fu, python plugins, etc. There's a ton of broken ones we'll have to fix. Something with the libarry not
#   being installed (these are broken upstream too, or due to my python setup maybe?)
#
# We flatten
#  icon
#  image
#  menu icons
#  mascot
# into one set of files, but we could break these out a bit; we could also figure out how to autogenerate the
# canvas shadow curves from the icon svg but this might be a bit far afield.
#

def setup_logging(verbose:bool=False, quiet:bool=False, logfile:Union[Path, str, None]=None) -> None:
    """
    Configure logging based on some flags.
    """
    # Setup Tqdm handler
    logger.setLevel(logging.DEBUG)
    h = TqdmLoggingHandler()
    if verbose:
        f = logging.Formatter('%(asctime)s %(module)-12s %(levelname)-8s %(message)s')
        h.setLevel(logging.DEBUG)
        h.setFormatter(f)
    elif quiet:
        f = logging.Formatter('%(levelname)-8s %(message)s')
        h.setLevel(logging.CRITICAL)
        h.setFormatter(f)
    else:
        f = logging.Formatter('%(levelname)-8s %(message)s')
        h.setLevel(logging.INFO)
        h.setFormatter(f)
    logger.addHandler(h)

    # setup logfile if specified
    if logfile:
        lf = logging.FileHandler(logfile)
        lf.setLevel(logging.DEBUG)
        lf.setFormatter(logging.Formatter('%(asctime)s %(module)-12s %(levelname)-8s %(message)s'))
        logger.addHandler(lf)


def parse_args(args: Sequence[str]) -> argparse.Namespace:
    """
    Parse command-line arguments.
    """
    parser = argparse.ArgumentParser(prog="heckimp",
                                     description="Process clean GnuImp source tree into rebranded fork.")

    parser.add_argument("-v", "--verbose", action="store_true", help="Show more verbosity but only a little.") # done
    parser.add_argument("-q", "--quiet", action="store_true", help="Show less innformation like not every filename.") # done
    parser.add_argument("--nogit", action="store_true", help="Don't do git stuff at the end.") # done

    # Disabling these for now because it complicates filename transforms. We can work out better filename transform logic later and support this.
    # For now you just need the checkout as 'gimp' and the output will be the filename.
    # parser.add_argument("-i", "--input", type=Path, default="gimp", help="Force input tree, default is original codebase name.")
    # parser.add_argument("-o", "--output", type=str, help="Force output tree. Default is the app's name.", default="")

    parser.add_argument("-c", "--config", type=Path, default=Path("heckimp-pika/pika.yaml"), help="Specify configuration file.", required=True) # done
    parser.add_argument("-p", "--paths", default=["."], nargs="+", help="Specify paths to look for images and stuff.") # done
    parser.add_argument("--dump", action="store_true", help="Don't excute, just dump configuration and replacement strings (for debugging purposes).") # done
    parser.add_argument("--log", type=str, default="", help="Write logs to specified file.") # done!

    parser.add_argument("--no-progress", dest='no_progress', action="store_true", help="Don't show progress bar.") # done

    # parser.add_argument("--simple", action="store_true", help="Simplify output (no progress bars and similar).") # fixme later [dependency loop]
    # parser.add_argument("--build", action="store_true", help="Also execute the build process with Meson and Ninja.") #Fixme maybe

    return parser.parse_args(args)


def heckimp(args: Sequence[str]):
    print(LOGO)
    pargs = parse_args(args[1:])
    setup_logging(pargs.verbose, pargs.quiet, pargs.log)

    inconfig = {}
    with pargs.config.open() as conffile:
        inconfig = yaml.safe_load(conffile)

    searchpaths = set([Path(path) for path in pargs.paths])
    searchpaths.add(pargs.config.parent)
    outdir = Path(inconfig['name'].lower())

    if 'shadow' in inconfig:
        shadow = find_file(inconfig['shadow'], searchpaths).read_text().replace('\n', ' ')
    else:
        shadow = ANNULUS

    tutorialurl = inconfig['url']
    if 'tutorialurl' in inconfig:
        tutorialurl = inconfig['tutorialurl']

    docsurl = inconfig['url']
    if 'docsurl' in inconfig:
        docsurl = inconfig['docsurl']

    contriburl = inconfig['url']
    if 'contriburl' in inconfig:
        contriburl = inconfig['contriburl']

    donateurl = inconfig['url']
    if 'donateurl' in inconfig:
        donateurl = inconfig['donateurl']

    bugurl = inconfig['url']
    if 'bugurl' in inconfig:
        bugurl = inconfig['bugurl']

    derivedconfig={
        # Derived variables from config
        "longname": inconfig['name'].upper(),
        "expandedname": inconfig['expansion'],
        "symname": inconfig['name'].capitalize(),
        "filename": inconfig['name'].lower(),
        "ufilename": inconfig['name'].upper(),
        "backdomain": reverse_domain(inconfig['url']),
        # URLs
        "baseurl": inconfig['url'],
        "tutorialurl": tutorialurl,
        "docsurl": docsurl,
        "contriburl": contriburl,
        "donateurl": donateurl,
        "bugurl": bugurl,
        #
        "copyright": inconfig['copyright'],
        "version": inconfig['version'],
        "resource_path": reverse_domain(inconfig['url']).replace('.', '/'),

        # Paths..
        "sourcetree": 'gimp',
        "outtree": outdir,
        "icon": inconfig['icon'],
        "splash": inconfig['splash'],
        "scalable": inconfig['scalable'],

        # heckimp internal variables
        "year": str(datetime.datetime.now().year),
        "us": "heckimp",
        "usversion": __version__,
        "uscopy": "Aldercone Studio",
        "shadow": shadow,
    }
    sourceroot = Path(derivedconfig["sourcetree"])

    visitfiles = set()
    seenfiles = set()
    targetpath = Path(derivedconfig["filename"])
    if targetpath.exists():
        visitfiles = set(targetpath.rglob('*'))


    # precompute string replacements
    for r in FILENAME_REGEXP:
        r.replacement = r.replacement.format(**derivedconfig)
    for r in CONTENTS_REGEXP:
        r.replacement = r.replacement.format(**derivedconfig)
    make_directories = [f.format(**derivedconfig) for f in MAKE_DIRECTORIES]

    if pargs.dump:
        # we can probably output regexes and stuff too but that seems a bit excessive right now.
        print("Dumping configuration:")
        print("## Input Configuration")
        pprint.pprint(inconfig)
        print("## Derived Configuration")
        pprint.pprint(derivedconfig)
        return 0


    # visit each file in the source tree,
    files = list(sourceroot.rglob('*'))

    if pargs.no_progress:
        loop = files
        print("Processing codebase")
    else:
        loop = tqdm.tqdm(files, desc="Processing input codebase", unit="file", dynamic_ncols=True, leave=False)

    for d in make_directories:
        logger.info(f"making path ahead of time: {d}")
        Path(d).mkdir(exist_ok=True, parents=True)
    for path in loop:
        if one_matches(FILENAME_EXCEPTIONS, str(path)):
            logger.info(f"skipping excluded path {path}")
            continue
        # possibly rename file,
        outpath = Path(apply_regexps(FILENAME_REGEXP, [], str(path)))
        seenfiles.add(outpath)
        logger.info(f"processing {outpath}")
        # write each file to the outtreea
        if path.is_dir():
            outpath.mkdir(exist_ok=True, parents=True)
            logger.info(f"making output directory -> {outpath}")
        else:
            #
            # fixme we should make a more functional way of doing all these special cases, like associating
            # the special regexs with a symbol that we match on instead so we can treat several of them the same
            # way
            #
            # do things to the contents
            m = ICON_REGEXP.match(str(path))
            n = MASCOT_REGEXP.match(str(path))
            # FIXME we should separate this special processing stuff into some separate functions, especially
            # the image resizing thing.
            if SPLASH_REGEXP.match(str(path)):
                logger.info(f"copying splash screen {derivedconfig['splash']}")
                # fixme instead of breaking this should copy a default heckimp splash
                splash = find_file(derivedconfig['splash'], searchpaths)
                shutil.copy(str(splash), str(outpath))
            elif SCALABLE_REGEXP.match(str(path)) or MASCOT_SCALABLE_REGEXP.match(str(path)):
                scalable = find_file(derivedconfig['scalable'], searchpaths)
                shutil.copy(str(scalable), str(outpath))
            elif LOGO_REGEXP.match(str(path)):
                iconpath = find_file(derivedconfig['icon'], searchpaths)
                icon = Image.open(str(iconpath))
                icon.thumbnail((256, 256), Image.ANTIALIAS)
                icon.save(str(outpath), 'PNG')
            elif m:
                # size icon for size specified in the path
                size = int(m.groups()[0])
                logger.info(f"resizing icon {derivedconfig['icon']} to {size}")
                iconpath = find_file(derivedconfig['icon'], searchpaths)
                icon = Image.open(str(iconpath))
                icon.thumbnail((size, size), Image.ANTIALIAS)
                icon.save(str(outpath), 'PNG')
            elif n:
                # this is identical to the above so we should reuse the code
                # size icon for size specified in the path
                size = int(n.groups()[0])
                logger.info(f"resizing icon {derivedconfig['icon']} to {size}")
                iconpath = find_file(derivedconfig['icon'], searchpaths)
                icon = Image.open(str(iconpath))
                icon.thumbnail((size, size), Image.ANTIALIAS)
                icon.save(str(outpath), 'PNG')
            elif one_matches(SKIP_PROCESSING, str(path)) or (path.suffix in PROBABLY_BINARY_EXTENSIONS) or is_binary(path):
                logger.info(f"skipping processing {outpath}")
                shutil.copy(str(path), str(outpath))
            else:
                # process each line of file to produce output
                process_text_file(CONTENTS_REGEXP, CONTENTS_EXCEPTIONS, path, outpath)
                # re-sort library definition files=
                if re.match(r".*\.def$", str(outpath)):
                    logger.info(f"sorting definition file {outpath}")
                    inf = outpath.read_text().split('\n')
                    outpath.write_text(inf[0]+('\n'.join(sorted(inf[1:], key=lambda d: d.strip()))))

    # post process patches
    # copy files:
    if 'files' in inconfig:
        for f in inconfig['files']:
            logger.info(f"Adding package-custom files: {f} -> {inconfig['files'][f]}")
            fpath = find_file(f, searchpaths)
            shutil.copy(str(fpath), os.path.join(derivedconfig['outtree'], inconfig['files'][f]))

    if not pargs.nogit:
        os.chdir(derivedconfig['outtree'])
        os.system('git init .')
        os.system('git add .')
        os.system(f"""git commit -m "Initial checkin of {derivedconfig['symname']} from {derivedconfig["us"]}" """)
        os.system(f"""git tag v{derivedconfig['version']}""")

    leftovers = visitfiles - seenfiles
    if leftovers:
        # print them out but we need to think about filtercing them too (exclude .git and upstream-documentation)
        ...
    print("Done.")
    return 0


def main():
    try:
        sys.exit(heckimp(sys.argv))
    except Exception as inst:
        # fixme log exception
        print(f"Exception! {inst}")
        print(traceback.format_exc())
        sys.exit(-1)



if __name__ == "__main__":
    main()
